
/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/errno.h>
#include "cx88_debug.h"
#include "cx88_capture.h"

using namespace std;

cx88_capture::cx88_capture(const string &id, int &error)
   : m_id(id),
     m_state(CX88_CAP_UNINITIALIZED)
{
   if (error)
   {
      return;
   }
   if ((error = pthread_mutex_init(&m_mutex, NULL)))
   {
      CX88ERR << "Unable to initialize capture lock" << endl;
      return;
   }
   m_state = CX88_CAP_MUTEX_CREATED;
   if ((error = pthread_cond_init(&m_avail, NULL)))
   {
      CX88ERR << "Unable to initialize capture signal" << endl;
      return;
   }
   m_state = CX88_CAP_CV_CREATED;
}

int cx88_capture::start(void)
{
   int error = initialize ();
   if (error)
   {
      return error;
   }
   if ((error = pthread_create(&m_thread, NULL, thread_routine, this)))
   {
      return error;
   }
   m_state = CX88_CAP_THREAD_CREATED;
   return error;
}

cx88_capture::~cx88_capture(void)
{
   if (m_state >= CX88_CAP_CV_CREATED)
   {
      pthread_cond_destroy(&m_avail);  
   }
   if (m_state >= CX88_CAP_MUTEX_CREATED)
   {
      pthread_mutex_destroy(&m_mutex);  
   }
}

void cx88_capture::stop(void)
{
   if (m_state == CX88_CAP_THREAD_CREATED)
   {
      pthread_mutex_lock(&m_mutex);
      m_state = CX88_CAP_THREAD_DYING;
      pthread_cond_signal(&m_avail);
      pthread_mutex_unlock(&m_mutex);
      pthread_join(m_thread, NULL);
      m_state = CX88_CAP_CV_CREATED;
   }
}

int cx88_capture::submit_workitem(cx88_capture_workitem &workitem)
{
   pthread_mutex_lock(&m_mutex);
   try
   {
      m_workqueue.push(workitem);
   }
   catch (...)
   {
      pthread_mutex_unlock(&m_mutex);
      CX88ERR << "Unable to submit item to work queue" << endl;
      return ENOMEM;  
   }
   pthread_cond_signal(&m_avail);
   pthread_mutex_unlock(&m_mutex);
   return 0; 
}

int cx88_capture::submit_buffer(void *buffer, size_t num_bytes, size_t packet_size)
{
   cx88_capture_workitem workitem;
   workitem.type = cx88_capture_workitem::CX88_CAPTURE_WORKITEM_BUFFER;
   workitem.buffer.buffer = buffer;
   workitem.buffer.num_bytes = num_bytes;
   workitem.buffer.packet_size = packet_size;
   return submit_workitem(workitem);
}

int cx88_capture::notify(uint32_t events, int error)
{
   cx88_capture_workitem workitem;
   workitem.type = cx88_capture_workitem::CX88_CAPTURE_WORKITEM_NOTIFICATION;
   workitem.notification.events = events;
   workitem.notification.error = error;
   return submit_workitem(workitem);
}

void *cx88_capture::thread_routine(void *arg)
{
   cx88_capture *cap = reinterpret_cast <cx88_capture*> (arg);
   pthread_mutex_lock(&(cap->m_mutex));
   while (cap->m_state != CX88_CAP_THREAD_DYING)
   {
      pthread_cond_wait(&(cap->m_avail), &(cap->m_mutex));
      while (!(cap->m_workqueue.empty()))
      {
         cx88_capture_workitem workitem = cap->m_workqueue.front();
         cap->m_workqueue.pop();
         pthread_mutex_unlock(&(cap->m_mutex));
         if (workitem.type == cx88_capture_workitem::CX88_CAPTURE_WORKITEM_BUFFER)
         {
            cap->process_buffer(workitem.buffer.buffer, workitem.buffer.num_bytes, workitem.buffer.packet_size);
         }
         else
         {
            cap->process_notification(workitem.notification.events, workitem.notification.error);
         }
         pthread_mutex_lock(&(cap->m_mutex));
      }
   }
   pthread_mutex_unlock(&(cap->m_mutex));
   cap->cleanup();
   return NULL;
}
