/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef __CX88_CAPTURE_H__
#define __CX88_CAPTURE_H__

#include <pthread.h>
#include <queue>
#include <vector>
#include <string>
#include "cx88_xml_config.h"
#include "cx88_capture.h"

class cx88_driver;

class cx88_capture
{
   public:

      cx88_capture(const std::string &id, int &error);
      
      virtual ~cx88_capture(void);
      
      int submit_buffer(void *buffer, size_t num_bytes, size_t packet_size);
      
      int start(void);
      
      void stop(void);
      
      const char *get_id(void)
      {
         return m_id.c_str();
      }

      virtual int set_driver(cx88_driver *driver)
      {
         return 0;
      }

      int notify(uint32_t events, int error);
      
   protected:

      virtual void process_notification(uint32_t events, int error) {}

      virtual void process_buffer(void *data, size_t num_bytes, size_t packet_size = 0) = 0;
      
      virtual int initialize(void) 
      {
         return 0;
      }
      
      virtual void flush(void) {}
      
      virtual void cleanup(void)
      {
         flush();
      }

      std::string m_id;

   private:

      typedef struct
      {
         enum
         {
            CX88_CAPTURE_WORKITEM_BUFFER,
            CX88_CAPTURE_WORKITEM_NOTIFICATION
         } type;
         union
         {
            struct
            {
               void *buffer;
               size_t num_bytes;
               size_t packet_size;
            } buffer;
            struct
            {
               uint32_t events;
               int error;
            } notification;
         };
      } cx88_capture_workitem;

      int submit_workitem(cx88_capture_workitem &workitem);

      std::queue<cx88_capture_workitem> m_workqueue;
      pthread_t m_thread;
      pthread_mutex_t m_mutex;
      pthread_cond_t m_avail;
      volatile enum
      {
         CX88_CAP_UNINITIALIZED,
         CX88_CAP_MUTEX_CREATED,
         CX88_CAP_CV_CREATED,
         CX88_CAP_THREAD_CREATED,
         CX88_CAP_THREAD_DYING
      } m_state;
      
      static void *thread_routine(void *arg);
};

typedef std::vector<cx88_capture*> capvec;

#endif
