/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <sys/errno.h>
#include "cx88_debug.h"
#include "cx88_driver.h"

using namespace std;

cx88_driver::cx88_driver(
   cx88_xml_config &cx88conf,
   tuner_config &tunerconf,
   const char *devnode,
   capvec &captures,
   int &error)
   : m_channel(cx88conf, tunerconf, error),
     m_tunerconf(tunerconf),
     m_devnode(devnode),
     m_stop(true),
     m_channel_changed(false),
     m_fd(-1),
     m_captures(captures),
     m_source_mask(0),
     m_lock_timeout_ms(1000),
     m_state(CX88_DRIVER_UNINITIALIZED)
{
   i2c_init();
   pthread_mutexattr_t attrs;
   if ((error = pthread_mutexattr_init(&attrs)))
   {
      CX88ERR << "Unable to initialize lock attributes" << endl;
      return;
   }
   if ((error = pthread_mutexattr_settype(&attrs, PTHREAD_MUTEX_RECURSIVE)))
   {
      CX88ERR << "Unable to set lock type to recursive" << endl;
      goto cleanup;
   }
   if ((error = pthread_mutex_init(&m_lock, &attrs)))
   {
      CX88ERR << "Unable to initialize driver lock" << endl;
      goto cleanup;
   }
   m_state = CX88_DRIVER_LOCK_CREATED;
   if ((error = pthread_cond_init(&m_signal, NULL)))
   {
      CX88ERR << "Unable to initialize driver event signal" << endl;
      goto cleanup;
   }
   m_state = CX88_DRIVER_SIGNAL_CREATED;
   if ((error = pthread_mutex_init(&m_capture_lock, &attrs)))
   {
      CX88ERR << "Unable to initialize capture lock" << endl;
      goto cleanup;
   }
   m_state = CX88_DRIVER_CAPTURE_LOCK_CREATED;
   m_params.completion = NULL;
   m_params.completion_arg = NULL;
   m_params.channel_event = NULL;
   m_params.channel_arg = NULL;
   m_params.duration_s = 0;
   m_params.lock_timeout_ms = m_lock_timeout_ms;
cleanup:
   pthread_mutexattr_destroy(&attrs);
}

cx88_driver::~cx88_driver(void)
{
   if (m_state >= CX88_DRIVER_CAPTURE_LOCK_CREATED)
   {
      pthread_mutex_destroy(&m_capture_lock);
   }
   if (m_state >= CX88_DRIVER_SIGNAL_CREATED)
   {
      pthread_cond_destroy(&m_signal);
   }
   if (m_state >= CX88_DRIVER_LOCK_CREATED)
   {
      pthread_mutex_destroy(&m_lock);
   }
   m_state = CX88_DRIVER_UNINITIALIZED;
}

int cx88_driver::change_channel(uint32_t lock_timeout_ms)
{
   m_lock_timeout_ms = lock_timeout_ms;
   if (!m_stop)
   {
      m_channel_changed = true;
      pthread_cond_signal(&m_signal);
   }
   return 0;
}

int cx88_driver::start(cx88_capture_params &params)
{
   cx88_driver_lock lock(this);
   int error = 0;
   if (m_stop)
   {
      m_params = params;
      m_stop = false;
      if ((error = pthread_create(&m_thread, NULL, thread_routine, this)))
      {
         CX88ERR << "Unable to create driver thread: " << strerror(error) << endl;
         m_stop = true;
         return error;
      }
   }
   return error;
}

int cx88_driver::stop(void)
{
   void *retval = NULL;
   {
      cx88_driver_lock lock(this);
      if (!m_stop)
      {
         m_stop = true;
         pthread_cond_signal(&m_signal);
      }
   }
   pthread_join(m_thread, &retval);
   return (reinterpret_cast <size_t> (retval));
}

void *cx88_driver::thread_routine(void *arg)
{
   cx88_driver *driver = reinterpret_cast<cx88_driver*>(arg);
   int error = driver->capture(driver->m_params);
   return ((void*)error);
}

int cx88_driver::start_captures(void)
{
   cx88_capture_lock lock(this);
   int error = 0;
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      if ((error = m_captures[i]->start()))
      {
         CX88ERR << "Error starting captures: " << strerror(error) << endl;
         break;
      }
   }
   return error;
}

int cx88_driver::add_capture(cx88_capture &cap)
{
   int error = 0;
   if (!m_stop)
   {
      error = cap.start();
   }
   if (!error)
   {
      try
      {
         cx88_capture_lock lock(this);
         m_captures.push_back(&cap);
      }
      catch (...)
      {
         cap.stop();
         error = ENOMEM;
      }
   }
   return error;
}

cx88_capture *cx88_driver::remove_capture(unsigned int index)
{
   cx88_capture *cap = NULL;
   {
      cx88_capture_lock lock(this);
      if (index >= m_captures.size())
      {
         return NULL;
      }
      cap = m_captures[index];
      m_captures.erase(m_captures.begin() + index);
   }
   cap->stop();
   return cap;
}

void cx88_driver::remove_capture(cx88_capture *cap)
{
   bool capture_found = false;
   {
      cx88_capture_lock lock(this);
      for (capvec::size_type i = 0; i < m_captures.size(); ++i)
      {
         if (m_captures[i] == cap)
         {
            m_captures.erase(m_captures.begin() + i);
            capture_found = true;
            break;
         }
      }
   }
   if (capture_found)
   {
      cap->stop();
   }
}

int cx88_driver::submit_buffer(void *buffer, size_t num_bytes, size_t packet_size)
{
   cx88_capture_lock lock(this);
   int error = 0;
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      if ((error = m_captures[i]->submit_buffer(buffer, num_bytes, packet_size)))
      {
         CX88ERR << "Error submitting buffer to captures: " << strerror(error) << endl;
         break;
      }
   }
   return error;
}

void cx88_driver::notify_captures(uint32_t event, int error)
{
   cx88_capture_lock lock(this);
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      m_captures[i]->notify(event, error);
   }
}

void cx88_driver::stop_captures(void)
{
   cx88_capture_lock lock(this);
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      m_captures[i]->stop();
   }
}

int cx88_driver::wait_for_event(uint32_t timeout_s)
{
   cx88_driver_lock lock(this);
   while (!m_stop && !m_channel_changed)
   {
      if (timeout_s == 0)
      {
         return pthread_cond_wait(&m_signal, &m_lock);
      }
      else
      {
         struct timespec systime;
         systime.tv_sec = time(NULL) + timeout_s;
         systime.tv_nsec = 0;
         return pthread_cond_timedwait(&m_signal, &m_lock, &systime);  
      }
   }
   return 0;
}

void cx88_driver::i2c_init(void)
{
   for (uint8_t i = 0; i < MAX_I2C_DEVS; ++i)
   {
      sprintf(m_i2c_devs[i].conf.in_driver, "iicdev");
      m_i2c_devs[i].conf.in_speed = CX88_I2C_SPEED_100KHZ;
      m_i2c_devs[i].conf.in_addr = 0;
      m_i2c_devs[i].dev = NULL;
      m_i2c_devs[i].driver = NULL;
      m_i2c_devs[i].bus = 0;
      m_i2c_devs[i].source_mask = CX88_SOURCE_DEFAULT;
   }
}

int cx88_driver::i2c_setup(void)
{
   cx88_driver_lock lock(this);
   i2c_init();
   int error = create_i2c_devices();
   for (uint8_t i = 0; i < CX23885_I2C_BUSES; ++i)
   {
      cx88_common_i2c_config config;
      config.in_bus_number = i;
      config.in_num_i2c_devices = 0;
      uint8_t config_indexes[CX88_COMMON_MAX_I2C_DEVICES];
      for (uint8_t j = 0; j < MAX_I2C_DEVS; ++j)
      {
         if ((m_i2c_devs[j].conf.in_addr != 0) && (m_i2c_devs[j].bus == i))
         {  
            config.devices[config.in_num_i2c_devices] = m_i2c_devs[j].conf;
            config_indexes[(config.in_num_i2c_devices)++] = j;
            if (config.in_num_i2c_devices == CX88_COMMON_MAX_I2C_DEVICES)
            {
               break;
            }
         }
      }
      if (config.in_num_i2c_devices == 0)
      {
         continue;
      }
      error = ioctl(m_fd, CX88_COMMON_I2C_CONFIG, &config);
      if (error)
      {
         CX88ERR << "Error issuing I2C CONFIG ioctl: " << strerror(errno) << endl;
         return errno;
      }
      for (uint8_t j = 0; j < config.in_num_i2c_devices; ++j)
      {
         char i2c_devnode[CX88_COMMON_MAX_DRIVER_NAME + 5];
         snprintf(i2c_devnode, CX88_COMMON_MAX_DRIVER_NAME, "/dev/%s%d", 
            m_i2c_devs[config_indexes[j]].conf.in_driver, config.devices[j].out_instance);
         m_i2c_devs[config_indexes[j]].dev = new tuner_devnode_device(m_tunerconf, i2c_devnode, error);
         if (m_i2c_devs[config_indexes[j]].dev == NULL)
         {
            error = ENOMEM;
         }
         if (error)
         {
            CX88ERR << "Unable to create control instance for peripheral device" << endl;
            i2c_teardown();
            break;
         }
      }
   }
   if (!error)
   {
      error = create_i2c_drivers();  
   }
   return error;
}

void cx88_driver::i2c_teardown(void)
{
   cx88_driver_lock lock(this);
   i2c_stop();
   destroy_i2c_drivers();
   destroy_i2c_devices();
   for (uint8_t i = 0; i < MAX_I2C_DEVS; ++i)
   {
      if (m_i2c_devs[i].dev != NULL)
      {
         delete m_i2c_devs[i].dev;
         m_i2c_devs[i].dev = NULL;
         m_i2c_devs[i].driver = NULL;
      }
   }
}

void cx88_driver::i2c_stop(void)
{
   cx88_driver_lock lock(this);
   for (uint8_t i = MAX_I2C_DEVS; i > 0; --i)
   {
      if ((m_i2c_devs[i - 1].driver != NULL) && (m_i2c_devs[i].source_mask & m_source_mask))
      {   
         m_i2c_devs[i - 1].driver->stop();
      }
   }
}

int cx88_driver::i2c_start(void)
{
   cx88_driver_lock lock(this);
   int error = 0;
   for (uint8_t i = 0; i < MAX_I2C_DEVS; ++i)
   {
      if (m_i2c_devs[i].driver != NULL)
      {   
         if (m_i2c_devs[i].source_mask & m_source_mask)
         {
            if ((error = m_i2c_devs[i].driver->start(m_lock_timeout_ms)))
            {
               CX88ERR << "Error starting peripheral device" << endl;
               return error;
            }
         }
         else
         {
            m_i2c_devs[i].driver->reset();
         }
      }
   }
   return 0;
}

int cx88_driver::command(vector<string> argv)
{
   CX88ERR << "Unrecognized command: " << argv[0].c_str() << endl;
   return 0;
}
