/*-
 * Copyright 2010 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/errno.h>
#include <signal.h>
#include "../cx88_debug.h"
#include "cx88_dvb_capture.h"

using namespace std;

const char* cx88_dvb_capture::m_dvb_devnames[] =
{
   "dvb/%s/demux0",
   "dvb/%s/dvr0",
   "dvb/%s/frontend0"
};

const struct cuse_methods cx88_dvb_capture::m_cuse_methods[] =
{
   {
      demux_open,
      demux_close,
      demux_read,
      NULL, // write
      demux_ioctl,
      demux_poll
   },
   {
      dvr_open,
      NULL, // close
      dvr_read,
      NULL, // write
      dvr_ioctl,
      dvr_poll
   },
   {
      frontend_open,
      frontend_close,
      NULL, // read
      NULL, // write
      frontend_ioctl,
      frontend_poll
   }
};

void cx88_dvb_capture::demux_feed_entry::insert(demux_feed_entry *prev, demux_feed_entry *next)
{
   m_prev = prev;
   m_next = next;
   if (prev != NULL)
   {
      prev->m_next = this;
   }
   if (next != NULL)
   {
      next->m_prev = this;
   }
}

void cx88_dvb_capture::demux_feed_entry::remove(void)
{
   if (m_next != NULL)
   {
      m_next->m_prev = m_prev;
   }
   if (m_prev != NULL)
   {
      m_prev->m_next = m_next;
   }
   m_prev = NULL;
   m_next = NULL;
}

static cx88_dvb_capture* cuse_owner = NULL;
static pthread_key_t buffer_key;
static pthread_key_t frontend_key;

static void hup_handler(int)
{
   cx88_dvb_buffer *buffer = (cx88_dvb_buffer*)pthread_getspecific(buffer_key);
   if (buffer != NULL)
   {
      buffer->cancel_transfer();
   }
   cx88_dvb_frontend *frontend = (cx88_dvb_frontend*)pthread_getspecific(frontend_key);
   if (frontend != NULL)
   {
      frontend->cancel_io();
   }
}

cx88_dvb_capture::cx88_dvb_capture(const string &id, const char *filename,
                                   int &error)
   : cx88_capture(id, error),
     m_num_io_threads(0),
     m_state(CX88_DVB_CAPTURE_UNINITIALIZED),
     m_dvr(10240 * 188, error),
     m_demux(m_dvr, error),
     m_frontend(error),
     m_feeds(NULL)
{
   error = (error ? error : pthread_mutex_init(&m_feed_lock, NULL));
   if (!error)
   {
      m_state = CX88_DVB_CAPTURE_FEED_LOCK_CREATED;
   }
   error = (error ? error : pthread_key_create(&buffer_key, NULL));
   if (!error)
   {
      m_state = CX88_DVB_CAPTURE_BUFFER_KEY_CREATED;
   }
   error = (error ? error : pthread_key_create(&frontend_key, NULL));
   if (!error)
   {
      m_state = CX88_DVB_CAPTURE_FRONTEND_KEY_CREATED;
   }
   if (!error && !__sync_bool_compare_and_swap(&cuse_owner, NULL, this))
   {
      error = EALREADY;
      CX88ERR << "DVB capture already in progress; only one allowed per device session" << endl;
   }
   error = (error ? error : cuse_init());
   if (!error)
   {
      m_state = CX88_DVB_CAPTURE_CUSE_INITIALIZED;
   }
   for (size_t i = 0; i < NUM_DVB_DEVICE_TYPES; ++i)
   {
      if (error)
      {
         m_cuse_devs[i] = NULL;
         continue;
      }
      m_cuse_devs[i] = cuse_dev_create(&m_cuse_methods[i], this, NULL, 0, 5, 0666, m_dvb_devnames[i], filename);
      if (m_cuse_devs[i] == NULL)
      {
         CX88ERR << "Unable to create DVB device node" << endl;
         error = ENOMEM;
      }
   }
   signal(SIGHUP, hup_handler);
   for (size_t i = 0; !error && (i < CX88_DVB_NUM_IO_THREADS); ++i)
   {
      error = pthread_create(&(m_cuse_threads[i]), NULL, cuse_thread_routine, NULL);
      if (!error)
      {
         ++m_num_io_threads;
      }
   }
}

cx88_dvb_capture::~cx88_dvb_capture(void)
{
   for (size_t i = 0; i < NUM_DVB_DEVICE_TYPES; ++i)
   {
      if (m_cuse_devs[i] != NULL)
      {
         cuse_dev_destroy(m_cuse_devs[i]);
         m_cuse_devs[i] = NULL;
      }
   }
   if (m_state >= CX88_DVB_CAPTURE_CUSE_INITIALIZED)
   {
      cuse_uninit();
   }
   __sync_bool_compare_and_swap(&cuse_owner, this, NULL);
   for (size_t i = 0; i < m_num_io_threads; ++i)
   {
      pthread_cancel(m_cuse_threads[i]);
      pthread_join(m_cuse_threads[i], NULL);
   }
   if (m_state >= CX88_DVB_CAPTURE_FRONTEND_KEY_CREATED)
   {
      pthread_key_delete(frontend_key);
   }
   if (m_state >= CX88_DVB_CAPTURE_BUFFER_KEY_CREATED)
   {
      pthread_key_delete(buffer_key);
   }
   if (m_state >= CX88_DVB_CAPTURE_FEED_LOCK_CREATED)
   {
      pthread_mutex_destroy(&m_feed_lock);
   }
   demux_feed_entry *entry = m_feeds;
   while (entry != NULL)
   {
      demux_feed_entry *next_entry = entry->get_next();
      delete entry;
      entry = next_entry;
   }
   m_feeds = NULL;
   m_state = CX88_DVB_CAPTURE_UNINITIALIZED;
}

void *cx88_dvb_capture::cuse_thread_routine(void *arg)
{
   while (cuse_wait_and_process() == 0);
   return NULL;
}

void cx88_dvb_capture::process_notification(uint32_t events, int error)
{
   if (events & cx88_driver::CX88_EVENT_CHANNEL)
   {
      m_frontend.channel_changed(error);
   }
}

int cx88_dvb_capture::dvr_open(struct cuse_dev *dev, int fflags)
{
   DIAGNOSTIC(CX88LOG << "cx88_dvb_capture::dvr_open()" << endl)
   return 0;
}

int cx88_dvb_capture::dvr_read(struct cuse_dev *dev, int fflags, void *user_ptr, int len)
{
   DIAGNOSTIC(CX88LOG << "cx88_dvb_capture::dvr_read()" << endl)
   cx88_dvb_capture *cap = (cx88_dvb_capture*)(cuse_dev_get_priv0(dev));
   if (cap == NULL)
   {
      return CUSE_ERR_INVALID;
   }
   size_t size = (size_t)len;
   pthread_setspecific(buffer_key, &(cap->m_dvr));
   int error = cap->m_dvr.read(user_ptr, fflags, size);
   pthread_setspecific(buffer_key, NULL);
   return (error ? error : (int)size);
}

int cx88_dvb_capture::dvr_poll(struct cuse_dev *dev, int fflags, int events)
{
   DIAGNOSTIC(CX88LOG << "cx88_dvb_capture::dvr_poll()" << endl)
   cx88_dvb_capture *cap = (cx88_dvb_capture*)(cuse_dev_get_priv0(dev));
   if (cap == NULL)
   {
      return 0;
   }
   return cap->m_dvr.poll(events);
}

int cx88_dvb_capture::dvr_ioctl(struct cuse_dev *dev, int fflags, unsigned long cmd, void *user_data)
{
   DIAGNOSTIC(CX88LOG << "cx88_dvb_capture::dvr_ioctl(" << hex << cmd << ")" << endl)
   cx88_dvb_capture *cap = (cx88_dvb_capture*)(cuse_dev_get_priv0(dev));
   if (cap == NULL)
   {
      return CUSE_ERR_INVALID;
   }
   if (cmd == DMX_SET_BUFFER_SIZE)
   {
      return cap->m_dvr.resize((size_t)user_data);
   }
   return CUSE_ERR_INVALID;
}

int cx88_dvb_capture::demux_open(struct cuse_dev *dev, int fflags)
{
   DIAGNOSTIC(CX88LOG << "cx88_dvb_capture::demux_open()" << endl)
   int error = 0;
   cx88_dvb_capture *cap = (cx88_dvb_capture*)(cuse_dev_get_priv0(dev));
   if (cap == NULL)
   {
      return CUSE_ERR_INVALID;
   }
   demux_feed_entry *entry = new demux_feed_entry(error);
   if (entry == NULL)
   {
      return CUSE_ERR_NO_MEMORY;
   }
   error = (error ? error : cap->m_frontend.start());
   if (error)
   {
      delete entry;
   }
   else
   {
      cuse_dev_set_per_file_handle(dev, entry);
      pthread_mutex_lock(&cap->m_feed_lock);
      entry->insert(NULL, cap->m_feeds);
      cap->m_feeds = entry;
      pthread_mutex_unlock(&cap->m_feed_lock);
   }
   return error;
}

int cx88_dvb_capture::demux_close(struct cuse_dev *dev, int fflags)
{
   DIAGNOSTIC(CX88LOG << "cx88_dvb_capture::demux_close()" << endl)
   cx88_dvb_capture *cap = (cx88_dvb_capture*)(cuse_dev_get_priv0(dev));
   demux_feed_entry *entry = (demux_feed_entry*)(cuse_dev_get_per_file_handle(dev));
   cuse_dev_set_per_file_handle(dev, NULL);
   if ((cap == NULL) || (entry == NULL))
   {
      return 0;
   }
   cap->m_demux.ioctl(entry->get_feed(), DMX_STOP, NULL);
   cap->m_frontend.stop();
   pthread_mutex_lock(&cap->m_feed_lock);
   if (entry == cap->m_feeds)
   {
      cap->m_feeds = entry->get_next();
   }
   entry->remove();
   pthread_mutex_unlock(&cap->m_feed_lock);
   delete entry;
   return 0;
}

int cx88_dvb_capture::demux_read(struct cuse_dev *dev, int fflags, void *user_ptr, int len)
{
   DIAGNOSTIC(CX88LOG << "cx88_dvb_capture::demux_read()" << endl)
   demux_feed_entry *entry = (demux_feed_entry*)(cuse_dev_get_per_file_handle(dev));
   if (entry == NULL)
   {
      return CUSE_ERR_FAULT;
   }
   size_t size = (size_t)len;
   pthread_setspecific(buffer_key, static_cast<cx88_dvb_buffer*>(&(entry->get_feed())));
   int error = entry->get_feed().read(user_ptr, fflags, size);
   pthread_setspecific(buffer_key, NULL);
   return (error ? error : (int)size);
}

int cx88_dvb_capture::demux_poll(struct cuse_dev *dev, int fflags, int events)
{
   DIAGNOSTIC(CX88LOG << "cx88_dvb_capture::demux_poll()" << endl)
   demux_feed_entry *entry = (demux_feed_entry*)(cuse_dev_get_per_file_handle(dev));
   if (entry == NULL)
   {
      return CUSE_ERR_FAULT;
   }
   return entry->get_feed().poll(events);
}

int cx88_dvb_capture::demux_ioctl(struct cuse_dev *dev, int fflags, unsigned long cmd, void *user_data)
{
   DIAGNOSTIC(CX88LOG << "cx88_dvb_capture::demux_ioctl(" << hex << cmd << ")" << endl)
   cx88_dvb_capture *cap = (cx88_dvb_capture*)(cuse_dev_get_priv0(dev));
   if (cap == NULL)
   {
      return CUSE_ERR_INVALID;
   }
   demux_feed_entry *entry = (demux_feed_entry*)(cuse_dev_get_per_file_handle(dev));
   if (entry == NULL)
   {
      return CUSE_ERR_FAULT;
   }
   return cap->m_demux.ioctl(entry->get_feed(), cmd, user_data);
}

int cx88_dvb_capture::frontend_open(struct cuse_dev *dev, int fflags)
{
   DIAGNOSTIC(CX88LOG << "cx88_dvb_capture::frontend_open()" << endl)
   cx88_dvb_capture *cap = (cx88_dvb_capture*)(cuse_dev_get_priv0(dev));
   if (cap == NULL)
   {
      return CUSE_ERR_INVALID;
   }
   if (fflags & CUSE_FFLAG_WRITE)
   {
      int error = cap->m_frontend.open();
      return (error ? error : cap->m_frontend.start());
   }
   return 0;
}

int cx88_dvb_capture::frontend_close(struct cuse_dev *dev, int fflags)
{
   DIAGNOSTIC(CX88LOG << "cx88_dvb_capture::frontend_close()" << endl)
   cx88_dvb_capture *cap = (cx88_dvb_capture*)(cuse_dev_get_priv0(dev));
   if (cap == NULL)
   {
      return CUSE_ERR_INVALID;
   }
   if (fflags & CUSE_FFLAG_WRITE)
   {
      cap->m_frontend.stop();
      cap->m_frontend.close();
   }
   return 0;
}

int cx88_dvb_capture::frontend_ioctl(struct cuse_dev *dev, int fflags, unsigned long cmd, void *user_data)
{
   DIAGNOSTIC(CX88LOG << "cx88_dvb_capture::frontend_ioctl(" << hex << cmd << ")" << endl)
   cx88_dvb_capture *cap = (cx88_dvb_capture*)(cuse_dev_get_priv0(dev));
   if (cap == NULL)
   {
      return CUSE_ERR_INVALID;
   }
   pthread_setspecific(frontend_key, &(cap->m_frontend));
   int error = cap->m_frontend.ioctl(fflags, cmd, user_data);
   pthread_setspecific(frontend_key, NULL);
   return error;
}

int cx88_dvb_capture::frontend_poll(struct cuse_dev *dev, int fflags, int events)
{
   DIAGNOSTIC(CX88LOG << "cx88_dvb_capture::frontend_poll()" << endl)
   cx88_dvb_capture *cap = (cx88_dvb_capture*)(cuse_dev_get_priv0(dev));
   if (cap == NULL)
   {
      return CUSE_ERR_INVALID;
   }
   return cap->m_frontend.poll(events);
}
