/*-
 * Copyright 2011 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/types.h>
extern "C"
{
   #include <cuse4bsd.h>
}
#include "../cx88_debug.h"
#include "cx88_dvb_demux.h"

using namespace std;

cx88_dvb_demux::cx88_dvb_demux(cx88_dvb_buffer &dvr, int &error)
   : m_dvr(dvr),
     m_feed_lock_created(false),
     m_last_packet_offset(0)
{
   error = (error ? error : pthread_mutex_init(&m_feed_lock, NULL));
   if (!error)
   {
      m_feed_lock_created = true;
   }
}

cx88_dvb_demux::~cx88_dvb_demux(void)
{
   if (m_feed_lock_created)
   {
      pthread_mutex_destroy(&m_feed_lock);
   }
}

int cx88_dvb_demux::add_feed(uint16_t pid, cx88_dvb_demux_feed &feed)
{
   int error = 0;
   try
   {
      demux_filter_map::iterator iter = m_filters.find(pid);
      if (iter == m_filters.end())
      {
         demux_pid_filter newfilter;
         iter = m_filters.insert(m_filters.begin(), pair<uint16_t, demux_pid_filter>(pid, newfilter));
      }
      iter->second.feeds.push_front(&feed);
   }
   catch (...)
   {
      error = CUSE_ERR_NO_MEMORY;
   }
   return error;
}

void cx88_dvb_demux::remove_feed(uint16_t pid, cx88_dvb_demux_feed &feed)
{
   demux_filter_map::iterator iter = m_filters.find(pid);
   if (iter != m_filters.end())
   {
      iter->second.feeds.remove(&feed);
      if (iter->second.feeds.empty())
      {
         m_filters.erase(iter);
      }
   }
}

int cx88_dvb_demux::add_feed(cx88_dvb_demux_feed &feed)
{
   cx88_dvb_demux_feed::pid_list::iterator iter = feed.get_pids().begin();
   int error = 0;
   while ((error == 0) && (iter != feed.get_pids().end()))
   {
      error = add_feed(*iter, feed);
      ++iter;
   }
   return error;
}

void cx88_dvb_demux::remove_feed(cx88_dvb_demux_feed &feed)
{
   cx88_dvb_demux_feed::pid_list::iterator iter = feed.get_pids().begin();
   while (iter != feed.get_pids().end())
   {
      remove_feed(*iter, feed);
      ++iter;
   }
}

uint8_t *cx88_dvb_demux::get_packet(uint8_t *&buffer, size_t &size)
{
   uint8_t *packet = buffer;
   if (m_last_packet_offset != 0)
   {
      if ((m_last_packet_offset + size) > 188)
      {
         if (packet[188 - m_last_packet_offset] == 0x47)
         {
            memcpy(&m_last_packet[m_last_packet_offset], packet, 188 - m_last_packet_offset);
            buffer = &buffer[188 - m_last_packet_offset];
            size -= (188 - m_last_packet_offset);
            m_last_packet_offset = 0;
            return m_last_packet;
         }
         else
         {
            m_last_packet_offset = 0;
         }
      }
      else
      {
         memcpy(&m_last_packet[m_last_packet_offset], packet, size);
         m_last_packet_offset += size;
         size = 0;
         return NULL;
      }
   }
   for (size_t i = 0; i < size; ++i)
   {
      if (packet[i] == 0x47)
      {
         if (((size > (i + 188)) && (packet[i + 188] == 0x47)))
         {
            size -= (i + 188);
            buffer = &buffer[i + 188];
            return &packet[i];
         }
         else if (size <= (i + 188))
         {
            memcpy(m_last_packet, &packet[i], size - i);
            m_last_packet_offset = size - i;
            break;
         }
      }
   }
   size = 0;
   return NULL;
}

int cx88_dvb_demux::submit_buffer(void *buffer, size_t size)
{
   uint8_t *packet_buffer = reinterpret_cast<uint8_t*>(buffer);
   size_t remaining_size = size;
   uint8_t *packet;
   pthread_mutex_lock(&m_feed_lock);
   while ((packet = get_packet(packet_buffer, remaining_size)) != NULL)
   {
      uint16_t pid = ((packet[1] & 0x1F) << 8) | packet[2];
      demux_filter_map::iterator iter = m_filters.find(pid);
      if (iter == m_filters.end())
      {
         continue;
      }
      size_t payload_offset = 188;
      if (packet[3] & 0x10)
      {
         uint8_t continuity = packet[3] & 0x0F;
         if (iter->second.initialized && (((iter->second.continuity + 1) & 0x0F) != continuity))
         {
            iter->second.invalidate = true;
         }
         iter->second.continuity = continuity;
         iter->second.initialized = true;
         payload_offset = 4;
         if (packet[3] & 0x20)
         {
            payload_offset += (packet[4] + 1);
            if ((payload_offset >= 188) || ((packet[4] > 0) && (packet[5] & 0x80)))
            {
               iter->second.invalidate = true;
               continue;
            }
         }
      }
      size_t crc_extent = 0;
      size_t crc_passed = 0;
      bool write_to_dvr = false;
      demux_feed_list::iterator feed_iter;
      for (feed_iter = iter->second.feeds.begin(); feed_iter != iter->second.feeds.end(); ++feed_iter)
      {
         if (iter->second.invalidate)
         {
            (*feed_iter)->invalidate();
         }
         cx88_dvb_demux_feed::demux_filter_result res = (*feed_iter)->filter_packet(packet, payload_offset, crc_extent, crc_passed);
         switch (res)
         {
            case cx88_dvb_demux_feed::DEMUX_FILTER_FINISHED:
            case cx88_dvb_demux_feed::DEMUX_FILTER_ACCEPT:
               write_to_dvr = (write_to_dvr ? true : ((*feed_iter)->get_output() == cx88_dvb_demux_feed::CX88_DVB_DEMUX_OUTPUT_DVR));
               if (res == cx88_dvb_demux_feed::DEMUX_FILTER_FINISHED)
               {
                  feed_iter = iter->second.feeds.erase(feed_iter);
                  --feed_iter;
               }
               break;
            default:
               break;
         }
      }
      iter->second.invalidate = false;
      if (write_to_dvr)
      {
         m_dvr.write(packet, 188);
      }
   }
   pthread_mutex_unlock(&m_feed_lock);
   return 0;
}

int cx88_dvb_demux::ioctl(cx88_dvb_demux_feed &feed, unsigned long cmd, void *user_data)
{
   int error = 0;
   pthread_mutex_lock(&m_feed_lock);
   switch (cmd)
   {
      case DMX_START:
         DIAGNOSTIC(CX88LOG << "cx88_dvb_demux: DMX_START" << endl);
         error = add_feed(feed);
         error = (error ? error : feed.start());
         break;
      case DMX_STOP:
         DIAGNOSTIC(CX88LOG << "cx88_dvb_demux: DMX_STOP" << endl);
         feed.stop();
         remove_feed(feed);
         break;
      case DMX_SET_FILTER:
      {
         DIAGNOSTIC(CX88LOG << "cx88_dvb_demux: DMX_SET_FILTER" << endl);
         struct dmx_sct_filter_params filter;
         error = cuse_copy_in(user_data, &filter, sizeof(filter));
         error = (error ? error : feed.set_filter(filter));
         if (!error && (filter.flags & DMX_IMMEDIATE_START))
         {
            error = add_feed(feed);
            error = (error ? error : feed.start());
         }
         break;
      }
      case DMX_SET_PES_FILTER:
      {
         DIAGNOSTIC(CX88LOG << "cx88_dvb_demux: DMX_SET_PES_FILTER" << endl);
         struct dmx_pes_filter_params filter;
         error = cuse_copy_in(user_data, &filter, sizeof(filter));
         error = (error ? error : feed.set_filter(filter));
         if (!error && (filter.flags & DMX_IMMEDIATE_START))
         {
            error = add_feed(feed);
            error = (error ? error : feed.start());
         }
         break;
      }
      case DMX_SET_BUFFER_SIZE:
         DIAGNOSTIC(CX88LOG << "cx88_dvb_demux: DMX_SET_BUFFER_SIZE" << endl);
         error = feed.resize((size_t)user_data);
         break;
      case DMX_ADD_PID:
      {
         DIAGNOSTIC(CX88LOG << "cx88_dvb_demux: DMX_ADD_PID" << endl);
         uint16_t pid;
         error = cuse_copy_in(user_data, &pid, sizeof(pid));
         error = (error ? error : feed.add_pid(pid));
         if (!error && (feed.get_state() == cx88_dvb_demux_feed::CX88_DVB_DEMUX_FILTER_RUNNING))
         {
            error = add_feed(pid, feed);
         }
         break;
      }
      case DMX_REMOVE_PID:
      {
         DIAGNOSTIC(CX88LOG << "cx88_dvb_demux: DMX_REMOVE_PID" << endl);
         uint16_t pid;
         error = cuse_copy_in(user_data, &pid, sizeof(pid));
         if (!error)
         {
            if (feed.get_state() == cx88_dvb_demux_feed::CX88_DVB_DEMUX_FILTER_RUNNING)
            {
               remove_feed(pid, feed);
            }
            feed.remove_pid(pid);
         }
         break;
      }
      default:
         DIAGNOSTIC(CX88LOG << "cx88_dvb_demux: unknown ioctl" << endl);
         return CUSE_ERR_INVALID;
   }
   pthread_mutex_unlock(&m_feed_lock);
   return error;
}
