#include <sys/types.h>
extern "C"
{
   #include <cuse4bsd.h>
}
#include "cx88_dvb_demux_feed.h"

cx88_dvb_demux_feed::cx88_dvb_demux_feed(int &error)
   : cx88_dvb_buffer(4096 * 2, error),
     m_secbuf_start(0),
     m_secbuf_pos(0),
     m_state(CX88_DVB_DEMUX_FILTER_UNCONFIGURED),
     m_type(CX88_DVB_DEMUX_FEED_NONE),
     m_flags(0),
     m_neq(false),
     m_reject(true)
{}

cx88_dvb_demux_feed::~cx88_dvb_demux_feed(void)
{}

int cx88_dvb_demux_feed::set_filter(uint16_t pid, uint32_t flags)
{
   stop();
   m_flags = flags;
   m_pids.clear();
   int error = add_pid(pid);
   if (error)
   {
      return error;
   }
   m_state = CX88_DVB_DEMUX_FILTER_STOPPED;
   return 0;
}

int cx88_dvb_demux_feed::set_filter(struct dmx_sct_filter_params &params)
{
   m_neq = false;
   for (uint8_t i = 0; i < DMX_FILTER_SIZE; ++i)
   {
      m_filter[i] = params.filter.filter[i];
      m_eq_mask[i] = params.filter.mask[i] & ~(params.filter.mode[i]);
      m_neq_mask[i] = params.filter.mask[i] & params.filter.mode[i];
      if (m_neq_mask[i])
      {
         m_neq = true;
      }
   }
   set_timeout(params.timeout);
   m_type = CX88_DVB_DEMUX_FEED_SECTION;
   m_output = CX88_DVB_DEMUX_OUTPUT_TSDEMUX;
   return set_filter(params.pid, params.flags);
}

int cx88_dvb_demux_feed::set_filter(struct dmx_pes_filter_params &params)
{
   switch (params.output)
   {
      case DMX_OUT_TAP:
         m_output = CX88_DVB_DEMUX_OUTPUT_PAYLOAD;
         break;
      case DMX_OUT_TSDEMUX_TAP:
         m_output = CX88_DVB_DEMUX_OUTPUT_TSDEMUX;
         break;
      case DMX_OUT_TS_TAP:
         m_output = CX88_DVB_DEMUX_OUTPUT_DVR;
         break;
      case DMX_OUT_DECODER:
         m_output = CX88_DVB_DEMUX_OUTPUT_NONE;
         break;
      default:
         return CUSE_ERR_INVALID;
   }
   m_type = CX88_DVB_DEMUX_FEED_PES;
   return set_filter(params.pid, params.flags);
}

int cx88_dvb_demux_feed::add_pid(uint16_t pid)
{
   if (!m_pids.empty() && (m_output != CX88_DVB_DEMUX_OUTPUT_TSDEMUX))
   {
      return CUSE_ERR_INVALID;
   }
   try
   {
      m_pids.insert(pid);
   }
   catch (...)
   {
      return CUSE_ERR_NO_MEMORY;
   }
   return 0;
}

int cx88_dvb_demux_feed::start(void)
{
   switch (m_state)
   {
      case CX88_DVB_DEMUX_FILTER_UNCONFIGURED:
         return CUSE_ERR_INVALID;
      case CX88_DVB_DEMUX_FILTER_RUNNING:
         reset();
      default:
         m_state = CX88_DVB_DEMUX_FILTER_RUNNING;
   }
   return 0;
}

void cx88_dvb_demux_feed::stop(void)
{
   if (m_state == CX88_DVB_DEMUX_FILTER_RUNNING)
   {
      reset();
      m_state = CX88_DVB_DEMUX_FILTER_STOPPED;
   }
}

void cx88_dvb_demux_feed::reset(void)
{
   invalidate();
   cx88_dvb_buffer::reset();
}

void cx88_dvb_demux_feed::invalidate(void)
{
   if (m_type == CX88_DVB_DEMUX_FEED_SECTION)
   {
      m_secbuf_start = 0;
      m_secbuf_pos = 0;
      m_reject = true;
   }
}

cx88_dvb_demux_feed::demux_filter_result
cx88_dvb_demux_feed::secbuf_write(uint8_t *packet,
                                  size_t offset,
                                  size_t size,
                                  size_t &crc_extent,
                                  size_t &crc_passed)
{
   if (m_reject || ((m_secbuf_pos + size) > sizeof(m_secbuf)))
   {
      m_reject = true;
      return DEMUX_FILTER_REJECT;
   }
   memcpy(m_secbuf + m_secbuf_pos, packet + offset, size);
   m_secbuf_pos += size;
   demux_filter_result res = DEMUX_FILTER_REJECT;
   size_t packet_offset = offset;
   while (m_secbuf_start < m_secbuf_pos)
   {
      uint16_t len = ((m_secbuf[m_secbuf_start + 1] & 0x0F) << 8) +
         m_secbuf[m_secbuf_start + 2] + 3;
      size_t section_end = m_secbuf_start + len;
      packet_offset += len;
      if (((len == 0) || (len > sizeof(m_secbuf))) ||
          ((m_flags & DMX_CHECK_CRC) && (crc_extent >= packet_offset) && (crc_passed < packet_offset)))
      {
         m_reject = true;
         res = DEMUX_FILTER_REJECT;
         break;
      }
      if (len > m_secbuf_pos)
      {
         res = DEMUX_FILTER_ACCEPT;
         break;
      }
      if ((m_flags & DMX_CHECK_CRC) && (crc_extent < packet_offset))
      {
         uint32_t crc = crc32(m_secbuf + m_secbuf_start, len - 4);
         uint32_t packet_crc = ((m_secbuf[section_end - 4] << 24) |
                                (m_secbuf[section_end - 3] << 16) |
                                (m_secbuf[section_end - 2] << 8)  |
                                 m_secbuf[section_end - 1]);
         if (packet_crc == crc)
         {
            crc_passed = packet_offset;
         }
         else
         {
            m_reject = true;
            res = DEMUX_FILTER_REJECT;
            break;
         }
         crc_extent = packet_offset;
      }
      size_t sec_offset = m_secbuf_start;
      bool neq_pass = !m_neq;
      res = DEMUX_FILTER_ACCEPT;
      for (size_t i = 0; i < DMX_FILTER_SIZE; ++i)
      {
         uint8_t diff = m_secbuf[sec_offset + i] ^ m_filter[i];
         if (m_eq_mask[i] & diff)
         {
            res = DEMUX_FILTER_REJECT;
            break;
         }
         if (m_neq_mask[i] & diff)
         {
            neq_pass = true;
         }
         if (i == 0)
         {
            sec_offset += 2;
         }
      }
      if (!neq_pass)
      {
         res = DEMUX_FILTER_REJECT;
      }
      if (res == DEMUX_FILTER_ACCEPT)
      {
         if (write(m_secbuf + m_secbuf_start, len) != 0)
         {
            m_reject = true;
            return DEMUX_FILTER_REJECT;
         }
         set_break();
         if (m_flags & DMX_ONESHOT)
         {
            return DEMUX_FILTER_FINISHED;
         }
      }
      if (m_secbuf[m_secbuf_start + 6] == m_secbuf[m_secbuf_start + 7])
      {
         m_secbuf_start = 0;
         m_secbuf_pos = 0;
         m_reject = false;
         break;
      }
      m_secbuf_start += len;
   }
   return res;
}

cx88_dvb_demux_feed::demux_filter_result cx88_dvb_demux_feed::filter_packet(
   uint8_t *packet,
   size_t payload_offset,
   size_t &crc_extent,
   size_t &crc_passed,
   size_t packet_size)
{
   switch (m_type)
   {
      case CX88_DVB_DEMUX_FEED_PES:
         switch (m_output)
         {
            case CX88_DVB_DEMUX_OUTPUT_TSDEMUX:
               return (write(packet, packet_size) == 0 ? 
                  DEMUX_FILTER_ACCEPT : DEMUX_FILTER_REJECT);
            case CX88_DVB_DEMUX_OUTPUT_PAYLOAD:
            {
               if (payload_offset >= packet_size)
               {
                  return DEMUX_FILTER_REJECT;
               }
               return (write(packet + payload_offset, packet_size - payload_offset) == 0 ?
                  DEMUX_FILTER_ACCEPT : DEMUX_FILTER_REJECT);
            }
            default:
               return DEMUX_FILTER_ACCEPT;
         }
      case CX88_DVB_DEMUX_FEED_SECTION:
      {
         if (payload_offset >= packet_size)
         {
            return DEMUX_FILTER_REJECT;
         }
         size_t section_end = packet_size;
         if (packet[1] & 0x40)
         {
            section_end = payload_offset + 1 + packet[payload_offset];
            ++payload_offset;
            if (section_end > packet_size)
            {
               invalidate();
               return DEMUX_FILTER_REJECT;
            }
         }
         demux_filter_result res = DEMUX_FILTER_REJECT;
         res = secbuf_write(packet, payload_offset, section_end - payload_offset, crc_extent, crc_passed);
         if ((res != DEMUX_FILTER_FINISHED) && (section_end < packet_size))
         {
            m_secbuf_start = 0;
            m_secbuf_pos = 0;
            m_reject = false;
            demux_filter_result res2 = secbuf_write(packet, section_end, packet_size - section_end, crc_extent, crc_passed);
            if (res2 != DEMUX_FILTER_REJECT)
            {
               res = res2;
            }
         }
         return res;
      }
      default:
         return DEMUX_FILTER_REJECT;
   }
}

// MPEG CRC algorithm adapted from NIST ATSC reference parser (NDRI)
uint32_t cx88_dvb_demux_feed::crc32(uint8_t *packet, size_t len)
{
   static const uint32_t polynomial = 0x04C11DB6;
   uint32_t shift_reg = 0xFFFFFFFF;

   for (size_t i = 0; i < len; ++i)
   {
      for (int j = 7; j >= 0; --j)
      {
         uint8_t data_bit = (packet[i] >> j) & 0x01;
         data_bit ^= ((shift_reg >> 31) & 0x01);
         shift_reg = (shift_reg << 1);
         if (data_bit == 1)
         {
            shift_reg = (shift_reg ^ polynomial) | 1 ;
         }
      }
   }
   return shift_reg;
}
