/*-
 * Copyright 2011 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
 
#include <sys/errno.h>
#include <sys/types.h>
#include <linux/dvb/version.h>
extern "C"
{
   #include <cuse4bsd.h>
}
#include "../cx88_debug.h"
#include "cx88_dvb_frontend.h"

using namespace std;

cx88_dvb_frontend::cx88_dvb_frontend(int &error)
   : m_driver(NULL),
     m_refcount(0),
     m_state(CX88_FRONTEND_UNINITIALIZED),
     m_signal_timestamp(0),
     m_oneshot(false),
     m_opened(false),
     m_canceled(false)
{
   memset(&m_info, 0, sizeof(m_info));
   memset(&m_properties, 0, sizeof(m_properties));
   memset(&m_param_cache, 0, sizeof(m_param_cache));
   memset(&m_signal, 0, sizeof(m_signal));
   error = (error ? error : pthread_mutex_init(&m_lock, NULL));
   if (!error)
   {
      m_state = CX88_FRONTEND_MUTEX_CREATED;
      error = pthread_cond_init(&m_event, NULL);
   }
   if (!error)
   {
      m_state = CX88_FRONTEND_CV_CREATED;
   }
}

cx88_dvb_frontend::~cx88_dvb_frontend(void)
{
   if (m_refcount > 0)
   {
      m_driver->stop();
   }
   if (m_state >= CX88_FRONTEND_CV_CREATED)
   {
      pthread_cond_destroy(&m_event);
   }
   if (m_state >= CX88_FRONTEND_MUTEX_CREATED)
   {
      pthread_mutex_destroy(&m_lock);
   }
   m_state = CX88_FRONTEND_UNINITIALIZED;
}

int cx88_dvb_frontend::set_driver(cx88_driver *driver)
{
   int error = 0;
   pthread_mutex_lock(&m_lock);
   m_driver = dynamic_cast<cx88_mpeg_driver*>(driver);
   if (m_driver == NULL)
   {
      error = EINVAL;
   }
   else
   {
      const char *name = m_driver->get_name();
      if (name != NULL)
      {
         strncpy(m_info.name, name, sizeof(m_info.name) - 1);
      }
      bool has_caps = false;
      uint32_t caps = m_driver->get_caps();
      m_info.caps = (fe_caps_t)0;
      if (caps & cx88_mpeg_driver::CX88_MPEG_CAP_ATSC)
      {
         m_info.type = FE_ATSC;
         m_info.frequency_min = 54000000;
         m_info.frequency_max = 890000000;
         m_info.frequency_stepsize = 62500;
         m_info.caps = (fe_caps_t)(FE_CAN_INVERSION_AUTO | FE_CAN_8VSB);
         has_caps = true;
      }
      if (caps & cx88_mpeg_driver::CX88_MPEG_CAP_QAM)
      {
         if (!has_caps)
         {
            m_info.type = FE_QAM;
            m_info.frequency_min = 54000000;
            m_info.frequency_max = 864000000;
            m_info.frequency_stepsize = 62500;
         }
         m_info.caps = (fe_caps_t)(m_info.caps | (FE_CAN_QAM_64 | FE_CAN_QAM_256 | FE_CAN_QAM_AUTO));
         has_caps = true;
      }
      if (caps & cx88_mpeg_driver::CX88_MPEG_CAP_OFDM)
      {
         if (!has_caps)
         {
            m_info.type = FE_OFDM;
            m_info.frequency_min = 177000000;
            m_info.frequency_max = 858000000;
            m_info.frequency_stepsize = 166667;
         }
         m_info.caps = (fe_caps_t)(m_info.caps | 
                        (FE_CAN_INVERSION_AUTO | FE_CAN_FEC_1_2 | FE_CAN_FEC_2_3 |
                         FE_CAN_FEC_3_4 | FE_CAN_FEC_4_5 | FE_CAN_FEC_5_6 |
                         FE_CAN_FEC_6_7 | FE_CAN_FEC_7_8 | FE_CAN_FEC_8_9 |
                         FE_CAN_FEC_AUTO | FE_CAN_TRANSMISSION_MODE_AUTO |
                         FE_CAN_QPSK | FE_CAN_QAM_16 | FE_CAN_QAM_64 | FE_CAN_QAM_AUTO |
                         FE_CAN_GUARD_INTERVAL_AUTO | FE_CAN_HIERARCHY_AUTO | FE_CAN_RECOVER));
         has_caps = true;
      }
      if (caps & cx88_mpeg_driver::CX88_MPEG_CAP_QPSK)
      {
         if (!has_caps)
         {
            m_info.type = FE_QPSK;
            m_info.frequency_min = 950000;
            m_info.frequency_max = 2150000;
            m_info.frequency_stepsize = 1000;
            m_info.symbol_rate_min = 1000000;
            m_info.symbol_rate_max = 45000000;
         }
         m_info.caps = (fe_caps_t)
                        (m_info.caps | (FE_CAN_INVERSION_AUTO | FE_CAN_FEC_1_2 | FE_CAN_FEC_2_3 |
                         FE_CAN_FEC_3_4 | FE_CAN_FEC_4_5 | FE_CAN_FEC_5_6 |
                         FE_CAN_FEC_6_7 | FE_CAN_FEC_7_8 | FE_CAN_FEC_AUTO |
                         FE_CAN_QPSK | FE_CAN_RECOVER));
         has_caps = true;
      }
   }
   pthread_mutex_unlock(&m_lock);
   return error;
}

int cx88_dvb_frontend::start(void)
{
   int error = 0;
   pthread_mutex_lock(&m_lock);
   if (m_refcount++ == 0)
   {
      if (m_driver->start() != 0)
      {
         error = CUSE_ERR_FAULT;
      }
   }
   pthread_mutex_unlock(&m_lock);
   return error;
}

void cx88_dvb_frontend::stop(void)
{
   pthread_mutex_lock(&m_lock);
   if ((m_refcount > 0) && (--m_refcount == 0))
   {
      cx88_driver *driver = m_driver;
      m_state = CX88_FRONTEND_STOPPED;
      m_events.clear();
      pthread_mutex_unlock(&m_lock);
      driver->stop();
   }
   else
   {
      pthread_mutex_unlock(&m_lock);
   }
}

void cx88_dvb_frontend::cancel_io(void)
{
   pthread_mutex_lock(&m_lock);
   m_canceled = true;
   pthread_cond_broadcast(&m_event);
   pthread_mutex_unlock(&m_lock);
}

int cx88_dvb_frontend::get_signal(void)
{
   int timestamp = time(NULL);
   if (abs(timestamp - m_signal_timestamp) >= 10)
   {
      m_signal_timestamp = timestamp;
      return m_driver->get_signal(m_signal);
   }
   return 0;
}

void cx88_dvb_frontend::channel_changed(int error)
{
   DIAGNOSTIC(CX88LOG << "cx88_dvb_frontend::channel_changed(" << error << ")" << endl);
   pthread_mutex_lock(&m_lock);
   if (!error && (m_state >= CX88_FRONTEND_CONFIGURED))
   {
      struct dvb_frontend_event event;
      event.parameters = m_params;
      if ((m_driver->get_signal(m_signal) != 0) || !m_signal.locked)
      {
         m_state = CX88_FRONTEND_ERROR;
      }
      else if (m_oneshot)
      {
         m_state = CX88_FRONTEND_TUNED;
      }
      else
      {
         event.status = (fe_status_t)(FE_HAS_SIGNAL | FE_HAS_CARRIER | 
                                      FE_HAS_SYNC | FE_HAS_VITERBI | FE_HAS_LOCK);
         try
         {
            m_events.push_front(event);
            m_state = CX88_FRONTEND_TUNED;
         }
         catch (...)
         {
            m_state = CX88_FRONTEND_ERROR;
         }
      }
      m_signal_timestamp = time(NULL);
   }
   else if (error)
   {
      m_signal.locked = false;
      m_state = CX88_FRONTEND_ERROR;
   }
   pthread_cond_broadcast(&m_event);
   pthread_mutex_unlock(&m_lock);
}

template <typename dvT, typename ltT, size_t n>
ltT cx88_dvb_frontend::dvb_to_libtuner(const ltT (&arr)[n], dvT index, int &error)
{
   size_t size = n;
   if ((size_t)index < size)
   {
      return arr[index];
   }
   error = (error ? error : CUSE_ERR_INVALID);
   return (ltT)0;
}

int cx88_dvb_frontend::set_params(dvb_channel channel, struct dvb_frontend_parameters &params)
{
   int error = 0;
   static const dvb_inversion_t inversion_table[] = {DVB_INVERSION_OFF, DVB_INVERSION_ON, DVB_INVERSION_AUTO};
   static const dvb_modulation_t modulation_table[] =
   {
      DVB_MOD_QPSK, DVB_MOD_QAM_16, DVB_MOD_QAM_32, DVB_MOD_QAM_64, DVB_MOD_QAM_128,
      DVB_MOD_QAM_256, DVB_MOD_QAM_AUTO, DVB_MOD_VSB_8, DVB_MOD_VSB_16, DVB_MOD_8PSK
   };
   static const dvb_fec_t fec_table[] =
   {
      DVB_FEC_NONE, DVB_FEC_1_2, DVB_FEC_2_3, DVB_FEC_3_4, DVB_FEC_4_5, DVB_FEC_5_6,
      DVB_FEC_6_7, DVB_FEC_7_8, DVB_FEC_8_9, DVB_FEC_AUTO, DVB_FEC_3_5, DVB_FEC_9_10
   };
   static const uint32_t bandwidth_table[] = {8000000, 7000000, 6000000};
   static const dvb_subcarrier_mode_t subcarriers_table[] =
   {
      DVB_OFDM_2K_SUBCARRIERS, DVB_OFDM_8K_SUBCARRIERS,
      DVB_OFDM_AUTO_SUBCARRIERS, DVB_OFDM_4K_SUBCARRIERS
   };
   static const dvb_guard_interval_t guard_interval_table[] =
   {
      DVB_OFDM_GUARD_1_32, DVB_OFDM_GUARD_1_16, DVB_OFDM_GUARD_1_8,
      DVB_OFDM_GUARD_1_4, DVB_OFDM_GUARD_AUTO
   };
   static const dvb_hierarchy_t hierarchy_table[] =
   {
      DVB_OFDM_HIERARCHY_NONE, DVB_OFDM_HIERARCHY_1, DVB_OFDM_HIERARCHY_2,
      DVB_OFDM_HIERARCHY_4, DVB_OFDM_HIERARCHY_AUTO
   };
   while (m_state == CX88_FRONTEND_CONFIGURED)
   {
      pthread_cond_wait(&m_event, &m_lock);
      if (m_canceled)
      {
         return CUSE_ERR_BUSY;
      }
   }
   channel.inversion = dvb_to_libtuner(inversion_table, params.inversion, error);
   if (error)
   {
      channel.inversion = DVB_INVERSION_AUTO;
      error = 0;
   }
   channel.frequency_hz = params.frequency;
   if (m_info.type == FE_QPSK)
   {
      channel.frequency_hz *= 1000;
   }
   switch (m_info.type)
   {
      case FE_ATSC:
         if (channel.bandwidth_hz == 0)
         {
            channel.bandwidth_hz = 6000000;
         }
         channel.modulation_settings.qam.symbol_rate = DVB_SYMBOL_RATE_AUTO;
         channel.modulation_settings.qam.code_rate = DVB_FEC_AUTO;
         channel.modulation = dvb_to_libtuner(modulation_table, params.u.vsb.modulation, error);
         break;
      case FE_QAM:
         if (channel.bandwidth_hz == 0)
         {
            channel.bandwidth_hz = 6000000;
         }
         channel.modulation_settings.qam.symbol_rate = params.u.qam.symbol_rate;
         channel.modulation = dvb_to_libtuner(modulation_table, params.u.qam.modulation, error);
         channel.modulation_settings.qam.code_rate = dvb_to_libtuner(fec_table, params.u.qam.fec_inner, error);
         break;
      case FE_OFDM:
         if (channel.bandwidth_hz == 0)
         {
            channel.bandwidth_hz = dvb_to_libtuner(bandwidth_table, params.u.ofdm.bandwidth, error);
         }
         channel.modulation = DVB_MOD_OFDM;
         channel.modulation_settings.ofdm.subcarrier_modulation = dvb_to_libtuner(modulation_table, params.u.ofdm.constellation, error);
         channel.modulation_settings.ofdm.high_prio_code_rate = dvb_to_libtuner(fec_table, params.u.ofdm.code_rate_HP, error);
         channel.modulation_settings.ofdm.low_prio_code_rate = dvb_to_libtuner(fec_table, params.u.ofdm.code_rate_LP, error);
         channel.modulation_settings.ofdm.subcarrier_mode = dvb_to_libtuner(subcarriers_table, params.u.ofdm.transmission_mode, error);
         channel.modulation_settings.ofdm.guard_interval = dvb_to_libtuner(guard_interval_table, params.u.ofdm.guard_interval, error);
         channel.modulation_settings.ofdm.hierarchy = dvb_to_libtuner(hierarchy_table, params.u.ofdm.hierarchy_information, error);
         break;
      case FE_QPSK:
         if (channel.modulation == DVB_MOD_UNKNOWN)
         {
            channel.modulation = DVB_MOD_QPSK;
         }
         channel.modulation_settings.psk.symbol_rate = params.u.qpsk.symbol_rate;
         channel.modulation_settings.psk.code_rate = dvb_to_libtuner(fec_table, params.u.qpsk.fec_inner, error);
         break;
      default:
         error = CUSE_ERR_FAULT;
         break;
   }
   if (m_driver->set_channel(channel) != 0)
   {
      error = CUSE_ERR_FAULT;
   }
   if (!error && !m_oneshot)
   {
      struct dvb_frontend_event event;
      event.parameters = params;
      event.status = (fe_status_t)0;
      try
      {
         m_events.push_front(event);
         pthread_cond_signal(&m_event);
      }
      catch (...)
      {
         error = CUSE_ERR_NO_MEMORY;
      }
   }
   if (!error)
   {
      m_params = params;
      m_state = CX88_FRONTEND_CONFIGURED;
   }
   else
   {
      m_state = CX88_FRONTEND_ERROR;
   }
   return error;
}

int cx88_dvb_frontend::ioctl(int fflags, unsigned long code, void *user_data)
{
   int error = 0;
   pthread_mutex_lock(&m_lock);
   if ((m_driver == NULL) || m_canceled)
   {
      error = CUSE_ERR_BUSY;
   }
   else
   {
      switch (code)
      {
         case FE_GET_INFO:
            DIAGNOSTIC(CX88LOG << "cx88_dvb_frontend: FE_GET_INFO" << endl)
            error = cuse_copy_out(&m_info, user_data, sizeof(m_info));
            break;
         case FE_READ_STATUS:
         {
            DIAGNOSTIC(CX88LOG << "cx88_dvb_frontend: FE_READ_STATUS" << endl)
            get_signal();
            int status = (m_signal.locked ? (FE_HAS_SIGNAL | FE_HAS_CARRIER | FE_HAS_LOCK) : 0);
            error = cuse_copy_out(&status, user_data, sizeof(status));
            break;
         }
         case FE_READ_BER:
         {
            DIAGNOSTIC(CX88LOG << "cx88_dvb_frontend: FE_READ_BER" << endl)
            get_signal();
            error = cuse_copy_out(&m_signal.ber, user_data, sizeof(m_signal.ber));
            break;
         }
         case FE_READ_SIGNAL_STRENGTH:
         {
            DIAGNOSTIC(CX88LOG << "cx88_dvb_frontend: FE_READ_SIGNAL_STRENGTH" << endl)
            get_signal();
            uint16_t strength = (uint16_t)((m_signal.strength * 0xFFFF) / 100);
            error = cuse_copy_out(&strength, user_data, sizeof(strength));
            break;
         }
         case FE_READ_SNR:
         {
            DIAGNOSTIC(CX88LOG << "cx88_dvb_frontend: FE_READ_SNR" << endl)
            get_signal();
            uint16_t snr = (uint16_t)((m_signal.snr * 0xFFFF) / 35);
            error = cuse_copy_out(&snr, user_data, sizeof(snr));
            break;
         }
         case FE_READ_UNCORRECTED_BLOCKS:
         {
            DIAGNOSTIC(CX88LOG << "cx88_dvb_frontend: FE_READ_UNCORRECTED_BLOCKS" << endl)
            get_signal();
            error = cuse_copy_out(&m_signal.uncorrected_blocks, user_data, sizeof(m_signal.uncorrected_blocks));
            break;
         }
         case FE_SET_FRONTEND:
         {
            DIAGNOSTIC(CX88LOG << "cx88_dvb_frontend: FE_SET_FRONTEND" << endl)
            struct dvb_frontend_parameters params;
            dvb_channel channel;
            memset(&channel, 0, sizeof(channel));
            if ((error = cuse_copy_in(user_data, &params, sizeof(params))))
            {
               break;
            }
            error = set_params(channel, params);
            break;
         }
         case FE_GET_FRONTEND:
            DIAGNOSTIC(CX88LOG << "cx88_dvb_frontend: FE_GET_FRONTEND" << endl)
            error = cuse_copy_out(&m_params, user_data, sizeof(m_params));
            break;
         case FE_GET_EVENT:
            DIAGNOSTIC(CX88LOG << "cx88_dvb_frontend: FE_GET_EVENT" << endl)
            while (m_events.empty())
            {
               if (fflags & CUSE_FFLAG_NONBLOCK)
               {
                  error = CUSE_ERR_WOULDBLOCK;
                  break;
               }
               else
               {
                  pthread_cond_wait(&m_event, &m_lock);
               }
               if (cuse_got_peer_signal() == 0)
               {
                  error = CUSE_ERR_SIGNAL;
                  break;
               }
               if (m_canceled)
               {
                  error = CUSE_ERR_BUSY;
                  break;
               }
            }
            if (!error)
            {
               struct dvb_frontend_event event = m_events.back();
               m_events.pop_back();
               error = cuse_copy_out(&event, user_data, sizeof(event));
            }
            break;
         case FE_SET_FRONTEND_TUNE_MODE:
            DIAGNOSTIC(CX88LOG << "cx88_dvb_frontend: FE_SET_FRONTEND_TUNE_MODE" << endl)
            m_oneshot = ((unsigned long)user_data & FE_TUNE_MODE_ONESHOT) ? true : false;
            break;
         case FE_SET_PROPERTY:
         {
            DIAGNOSTIC(CX88LOG << "cx88_dvb_frontend: FE_SET_PROPERTY" << endl)
            struct dtv_properties arg;
            if ((error = cuse_copy_in(user_data, &arg, sizeof(arg))))
            {
               break;
            }
            if (arg.num > DTV_IOCTL_MAX_MSGS)
            {
               error = CUSE_ERR_INVALID;
               break;
            }
            struct dtv_property *props = new struct dtv_property[arg.num];
            if (props == NULL)
            {
               error = CUSE_ERR_NO_MEMORY;
               break;
            }
            error = (error ? error : cuse_copy_in(arg.props, props, arg.num * sizeof(struct dtv_property)));
            error = (error ? error : set_properties(props, arg.num));
            error = (error ? error : cuse_copy_out(props, arg.props, arg.num * sizeof(struct dtv_property)));
            delete[] props;
            break;
         }
         case FE_GET_PROPERTY:
         {
            DIAGNOSTIC(CX88LOG << "cx88_dvb_frontend: FE_GET_PROPERTY" << endl)
            struct dtv_properties arg;
            if ((error = cuse_copy_in(user_data, &arg, sizeof(arg))))
            {
               break;
            }
            if (arg.num > DTV_IOCTL_MAX_MSGS)
            {
               error = CUSE_ERR_INVALID;
               break;
            }
            struct dtv_property *props = new struct dtv_property[arg.num];
            if (props == NULL)
            {
               error = CUSE_ERR_NO_MEMORY;
               break;
            }
            error = (error ? error : cuse_copy_in(arg.props, props, arg.num * sizeof(struct dtv_property)));
            error = (error ? error : get_properties(props, arg.num));
            error = (error ? error : cuse_copy_out(props, arg.props, arg.num * sizeof(struct dtv_property)));
            delete[] props;
            break;
         }
         default:
            DIAGNOSTIC(CX88LOG << "cx88_dvb_frontend: unknown ioctl " << hex << code << endl)
            error = CUSE_ERR_INVALID;
            break;
      }
   }
   pthread_mutex_unlock(&m_lock);
   DIAGNOSTIC(CX88LOG << "cx88_dvb_frontend: ioctl returning " << error << endl);
   return error;
}

int cx88_dvb_frontend::set_properties(struct dtv_property *props, size_t num_props)
{
   int error = 0;
   for (size_t i = 0; i < num_props; ++i)
   {
      props[i].result = 0;
      switch(props[i].cmd)
      {
         case DTV_CLEAR:
            memset(&m_properties, 0, sizeof(m_properties));
            memset(&m_param_cache, 0, sizeof(m_param_cache));
            break;
         case DTV_TUNE:
            error = set_params(m_properties, m_param_cache);
            if (error)
            {
               props[i].result = EINVAL;
            }
            break;
         case DTV_FREQUENCY:
            m_param_cache.frequency = props[i].u.data;
            break;
         case DTV_MODULATION:
            switch (m_info.type)
            {
               case FE_ATSC:
                  m_param_cache.u.vsb.modulation = (fe_modulation_t)(props[i].u.data);
                  break;
               case FE_QAM:
                  m_param_cache.u.qam.modulation = (fe_modulation_t)(props[i].u.data);
                  break;
               case FE_OFDM:
                  m_param_cache.u.ofdm.constellation = (fe_modulation_t)(props[i].u.data);
                  break;
               case FE_QPSK:
                  switch (props[i].u.data)
                  {
                     case PSK_8:
                        m_properties.modulation = DVB_MOD_8PSK;
                        break;
                     case APSK_16:
                        m_properties.modulation = DVB_MOD_16PSK;
                        break;
                     default:
                        break;
                  }
                  break;
               default:
                  break;
            }
            break;
         case DTV_BANDWIDTH_HZ:
            m_properties.bandwidth_hz = props[i].u.data;
            if (m_info.type == FE_OFDM)
            {
               switch (props[i].u.data)
               {
                  case 6000000:
                     m_param_cache.u.ofdm.bandwidth = BANDWIDTH_6_MHZ;
                     break;
                  case 7000000:
                     m_param_cache.u.ofdm.bandwidth = BANDWIDTH_7_MHZ;
                     break;
                  case 8000000:
                     m_param_cache.u.ofdm.bandwidth = BANDWIDTH_8_MHZ;
                     break;
                  default:
                     m_param_cache.u.ofdm.bandwidth = BANDWIDTH_AUTO;
                     break;
               }
            }
            break;
         case DTV_INVERSION:
            m_param_cache.inversion = (fe_spectral_inversion_t)(props[i].u.data);
            break;
         case DTV_SYMBOL_RATE:
            switch (m_info.type)
            {
               case FE_QAM:
                  m_param_cache.u.qam.symbol_rate = props[i].u.data;
                  break;
               case FE_QPSK:
                  m_param_cache.u.qpsk.symbol_rate = props[i].u.data;
               default:
                  break;
            }
            break;
         case DTV_INNER_FEC:
            switch (m_info.type)
            {
               case FE_QAM:
                  m_param_cache.u.qam.fec_inner = (fe_code_rate_t)(props[i].u.data);
                  break;
               case FE_QPSK:
                  m_param_cache.u.qpsk.fec_inner = (fe_code_rate_t)(props[i].u.data);
               default:
                  break;  
            }
            break;
         case DTV_PILOT:
         {
            static const dvb_pilot_t pilot_table[] = {DVB_PILOT_ON, DVB_PILOT_OFF, DVB_PILOT_AUTO};
            if (m_info.type == FE_QPSK)
            {
               m_properties.modulation_settings.psk.pilot = dvb_to_libtuner(pilot_table, props[i].u.data, error);
            }
            break;
         }
         case DTV_ROLLOFF:
         {
            static const dvb_rolloff_t rolloff_table[] = {DVB_ROLLOFF_0_35, DVB_ROLLOFF_0_20, DVB_ROLLOFF_0_25, DVB_ROLLOFF_AUTO};
            if (m_info.type == FE_QPSK)
            {
               m_properties.modulation_settings.psk.rolloff_factor = dvb_to_libtuner(rolloff_table, props[i].u.data, error);
            }
            break;
         }
         case DTV_CODE_RATE_HP:
            if (m_info.type == FE_OFDM)
            {
               m_param_cache.u.ofdm.code_rate_HP = (fe_code_rate_t)(props[i].u.data);
            }
            break;
         case DTV_CODE_RATE_LP:
            if (m_info.type == FE_OFDM)
            {
               m_param_cache.u.ofdm.code_rate_LP = (fe_code_rate_t)(props[i].u.data);
            }
            break;
         case DTV_GUARD_INTERVAL:
            if (m_info.type == FE_OFDM)
            {
               m_param_cache.u.ofdm.guard_interval = (fe_guard_interval_t)(props[i].u.data);
            }
            break;
         case DTV_TRANSMISSION_MODE:
            if (m_info.type == FE_OFDM)
            {
               m_param_cache.u.ofdm.transmission_mode = (fe_transmit_mode_t)(props[i].u.data);
            }
            break;
         case DTV_HIERARCHY:
            if (m_info.type == FE_OFDM)
            {
               m_param_cache.u.ofdm.hierarchy_information = (fe_hierarchy_t)(props[i].u.data);
            }
            break;
         default:
            break;
      }
   }
   return error;
}

int cx88_dvb_frontend::get_properties(struct dtv_property *props, size_t num_props)
{
   int error = 0;
   for (size_t i = 0; i < num_props; ++i)
   {
      props[i].result = 0;
      props[i].u.data = 0;
      switch(props[i].cmd)
      {
         case DTV_FREQUENCY:
            props[i].u.data = m_param_cache.frequency;
            break;
         case DTV_MODULATION:
            switch (m_info.type)
            {
               case FE_ATSC:
                  props[i].u.data = m_param_cache.u.vsb.modulation;
                  break;
               case FE_QAM:
                  props[i].u.data = m_param_cache.u.qam.modulation;
                  break;
               case FE_OFDM:
                  props[i].u.data = m_param_cache.u.ofdm.constellation;
                  break;
               case FE_QPSK:
                  switch (m_properties.modulation)
                  {
                     case DVB_MOD_8PSK:
                        props[i].u.data = PSK_8;
                        break;
                     case DVB_MOD_16PSK:
                        props[i].u.data = APSK_16;
                        break;
                     default:
                        props[i].u.data = QPSK;
                        break;
                  }
                  break;
               default:
                  break;
            }
            break;
         case DTV_BANDWIDTH_HZ:
            props[i].u.data = m_properties.bandwidth_hz;
            break;
         case DTV_INVERSION:
            props[i].u.data = m_param_cache.inversion;
            break;
         case DTV_SYMBOL_RATE:
            switch (m_info.type)
            {
               case FE_QAM:
                  props[i].u.data = m_param_cache.u.qam.symbol_rate;
                  break;
               case FE_QPSK:
                  props[i].u.data = m_param_cache.u.qpsk.symbol_rate;
               default:
                  props[i].u.data = 0;
                  break;
            }
            break;
         case DTV_INNER_FEC:
            switch (m_info.type)
            {
               case FE_QAM:
                  props[i].u.data = m_param_cache.u.qam.fec_inner;
                  break;
               case FE_QPSK:
                  props[i].u.data = m_param_cache.u.qpsk.fec_inner;
                  break;
               default:
                  props[i].u.data = FEC_NONE;
                  break;  
            }
            break;
         case DTV_PILOT:
         {
            static const fe_pilot_t pilot_table[] = {PILOT_AUTO, PILOT_OFF, PILOT_ON};
            if (m_info.type == FE_QPSK)
            {
               props[i].u.data = dvb_to_libtuner(pilot_table, m_properties.modulation_settings.psk.pilot, error);
            }
            else
            {
               props[i].u.data = PILOT_AUTO;
            }
            break;
         }
         case DTV_ROLLOFF:
         {
            static const fe_rolloff_t rolloff_table[] = {ROLLOFF_AUTO, ROLLOFF_20, ROLLOFF_25, ROLLOFF_35};
            if (m_info.type == FE_QPSK)
            {
               props[i].u.data = dvb_to_libtuner(rolloff_table, m_properties.modulation_settings.psk.rolloff_factor, error);
            }
            else
            {
               props[i].u.data = ROLLOFF_AUTO;
            }
            break;
         }
         case DTV_DELIVERY_SYSTEM:
            switch (m_info.type)
            {
               case FE_ATSC:
                  switch (m_properties.modulation)
                  {
                     case DVB_MOD_UNKNOWN:
                     case DVB_MOD_VSB_8:
                     case DVB_MOD_VSB_16:
                        props[i].u.data = SYS_ATSC;
                        break;
                     default:
                        props[i].u.data = SYS_DVBC_ANNEX_B;
                        break;
                  }
                  break;
               case FE_QAM:
                  props[i].u.data = SYS_DVBC_ANNEX_AC;
                  break;
               case FE_OFDM:
                  props[i].u.data = SYS_DVBT;
                  break;
               case FE_QPSK:
                  props[i].u.data = SYS_DVBS;
                  break;
            }
            break;
         case DTV_API_VERSION:
            props[i].u.data = (DVB_API_VERSION << 8) | DVB_API_VERSION_MINOR;
            break;
         case DTV_CODE_RATE_HP:
            if (m_info.type == FE_OFDM)
            {
               props[i].u.data = m_param_cache.u.ofdm.code_rate_HP;
            }
            break;
         case DTV_CODE_RATE_LP:
            if (m_info.type == FE_OFDM)
            {
               props[i].u.data = m_param_cache.u.ofdm.code_rate_LP;
            }
            break;
         case DTV_GUARD_INTERVAL:
            if (m_info.type == FE_OFDM)
            {
               props[i].u.data = m_param_cache.u.ofdm.guard_interval;
            }
            break;
         case DTV_TRANSMISSION_MODE:
            if (m_info.type == FE_OFDM)
            {
               props[i].u.data = m_param_cache.u.ofdm.transmission_mode;
            }
            break;
         case DTV_HIERARCHY:
            if (m_info.type == FE_OFDM)
            {
               props[i].u.data = m_param_cache.u.ofdm.hierarchy_information;
            }
            break;
         default:
            break;
      }
   }
   return error;
}

int cx88_dvb_frontend::open(void)
{
   int error = CUSE_ERR_BUSY;
   pthread_mutex_lock(&m_lock);
   if (!m_opened)
   {
      error = 0;
      m_opened = true;
      m_oneshot = false;
      m_canceled = false;
      m_events.clear();
      memset(&m_properties, 0, sizeof(m_properties));
      memset(&m_param_cache, 0, sizeof(m_param_cache));
   }
   pthread_mutex_unlock(&m_lock);
   return error;
}

void cx88_dvb_frontend::close(void)
{
   pthread_mutex_lock(&m_lock);
   m_opened = false;
   pthread_mutex_unlock(&m_lock);
}

int cx88_dvb_frontend::poll(int events)
{
   int mask = 0;
   pthread_mutex_lock(&m_lock);
   if (!m_events.empty())
   {
      mask = events & CUSE_POLL_READ;
   }
   pthread_mutex_unlock(&m_lock);
   return mask;
}
